/* MMTestFuncs.c */

#include "MMTestFuncs.h"

/* Unique identifier styles. Use whatever suits -- just make sure it's universally unique. */

/* UUID style (Hint: use uuidgen or CFUUIDRef to generate a uuid string. */
//const CFStringRef kPermanentLibraryIDFileSystem = CFSTR( "049C2566-F179-4F0C-80D6-3F892DF52D06" );
//const CFStringRef kPermanentLibraryIDFiles = CFSTR( "7C38C3FC-C296-457B-872B-D6648E1733C5" );
//const CFStringRef kPermanentLibraryIDFolders = CFSTR( "D6BCB177-A86B-4F80-922C-9DD0196C1BB9" );
//const CFStringRef kPermanentLibraryIDDisks = CFSTR( "EF073E7C-7E57-4902-B3C6-0539922B8DE3" );

/* Bundle ID style */
const CFStringRef kPermanentLibraryIDFileSystem = CFSTR( "com.example.library.FileSystem" );
const CFStringRef kPermanentLibraryIDFiles = CFSTR( "com.example.library.Files" );
const CFStringRef kPermanentLibraryIDFolders = CFSTR( "com.example.library.Folders" );
const CFStringRef kPermanentLibraryIDDisks = CFSTR( "com.example.library.Disks" );
const CFStringRef kPermanentLibraryIDDisksSub10 = CFSTR( "com.example.library.DisksSub10" );
const CFStringRef kPermanentLibraryIDDisksSub20 = CFSTR( "com.example.library.DisksSub20" );
const CFStringRef kPermanentLibraryIDDisksSub21 = CFSTR( "com.example.library.DisksSub21" );
const CFStringRef kPermanentLibraryIDEmpty = CFSTR( "com.example.empty.library" );

const CFStringRef kPermanentSearchConfigUserPlusFiles = CFSTR( "com.example.config.UserPlusFiles" );


static CFStringRef AEPrint( const AEDesc *theDesc );
static void ShowSomeValues( const void *theValue, void *ioContext );
static MMErrorRef CreateEmptyLibrary( CFAllocatorRef theAllocator );
//static
MMErrorRef CreateNewLibraryAndFolder( CFAllocatorRef theAllocator, CFStringRef theParentID, CFStringRef thePermanentID, CFStringRef theFolderName, CFStringRef theLibraryTitle, CFStringRef theNotes );
//static
CFStringRef CurrentImproperlyFormattedISODate( CFAllocatorRef theAllocator );
static void GetTitles( const void *theValue, void *ioContext );
//static
void ShowUUIDArray( const void *theValue, void *ioContext );
//static
void ShowStringArray( const void *theValue, void *ioContext );


CFStringRef CurrentImproperlyFormattedISODate( CFAllocatorRef theAllocator ) {
	
	CFStringRef isot = NULL;
	CFAbsoluteTime abst = CFAbsoluteTimeGetCurrent();
	CFTimeZoneRef tz = CFTimeZoneCopyDefault();
	CFGregorianDate gdate = CFAbsoluteTimeGetGregorianDate( abst, tz );
	
	isot = CFStringCreateWithFormat( theAllocator, NULL, CFSTR( "%.4d-%.2d-%.2d %.2d:%.2d:%0.2d" ), gdate.year, gdate.month, gdate.day, gdate.hour, gdate.minute, (int)gdate.second );
	if( tz != NULL ) {
		CFRelease( tz );
	}
	
	return isot;
}


CFStringRef AEPrint( const AEDesc *theDesc ) {
	
	CFStringRef rslt = NULL;
	OSStatus err = paramErr;
	Handle handle = NULL;
	err = AEPrintDescToHandle( theDesc, &handle );
	if( err == noErr ) {
		
		int size = (int)GetHandleSize( handle );
		
		rslt = CFStringCreateWithFormat( kCFAllocatorDefault, NULL, CFSTR( "%.*s\n" ), size, *handle );
		DisposeHandle( handle );
	}
	
	return rslt;
}


void ShowUUIDArray( const void *theValue, void *ioContext ) {
	
	MMUUIDRef uuid = theValue;
	CFMutableStringRef mute = (CFMutableStringRef)ioContext;
	
	if( mute != NULL ) {
		CFStringAppendFormat( mute, NULL, CFSTR( "\tUUID: %@" ), uuid );
		CFStringAppend( mute, CFSTR( "\n" ) );
	}
}


void ShowStringArray( const void *theValue, void *ioContext ) {
	
	CFStringRef cfst = theValue;
	CFMutableStringRef mute = (CFMutableStringRef)ioContext;
	
	if( mute != NULL ) {
		CFStringAppendFormat( mute, NULL, CFSTR( "\t%@" ), cfst );
		CFStringAppend( mute, CFSTR( "\n" ) );
	}
}


void ShowSomeValues( const void *theValue, void *ioContext ) {
	
	MMUUIDRef library_id = theValue;
	CFMutableStringRef mute = (CFMutableStringRef)ioContext;
	
	if( mute != NULL ) {
		
		CFStringRef title = MMLibraryPropertyGetValue( kCFAllocatorDefault, library_id, kMMKeyTitle, NULL );
		CFStringRef notes = MMLibraryPropertyGetValue( kCFAllocatorDefault, library_id, kMMKeyNotes, NULL );
		
		if( title != NULL ) {
			CFStringAppendFormat( mute, NULL, CFSTR( "Library ID '%@' has title: \"%@\"" ), library_id, title );
		}
		if( notes != NULL ) {
			CFStringAppendFormat( mute, NULL, CFSTR( " and notes: \"%@\"" ), notes );
		}
		CFStringAppend( mute, CFSTR( "\n" ) );
	}
}


void GetTitles( const void *theValue, void *ioContext ) {
	
	MMUUIDRef library_id = theValue;
	CFMutableStringRef mute = (CFMutableStringRef)ioContext;
		
	if( mute != NULL ) {
		
		CFStringRef title = (CFStringRef)MMLibraryPropertyGetValue( kCFAllocatorDefault, library_id, kMMKeyTitle, NULL );
		
		if( title != NULL ) {
			CFStringAppendFormat( mute, NULL, CFSTR( "Found library ID '%@'" ), library_id );
			CFStringAppendFormat( mute, NULL, CFSTR( " with title: \"%@\"\n" ), title );
		}
		else {
			CFStringAppendFormat( mute, NULL, CFSTR( "Library ID '%@' not found\n" ), library_id );
		}
	}
}


MMErrorRef CreateEmptyLibrary( CFAllocatorRef theAllocator ) {
	
	MMErrorRef eref = NULL;
	MMUUIDRef ldid = MMLibraryDescriptionCreateMinimal( theAllocator, kMMLocalLibraryID, kPermanentLibraryIDEmpty, CFSTR( "Empty Local Library" ), &eref );
	
	if( ldid != NULL ) {
		eref = MMDocumentRegisterLibrary( theAllocator, ldid, NULL );
		CFRelease( ldid );
	}
	
	return eref;
}


MMErrorRef CreateNewLibraryAndFolder( CFAllocatorRef theAllocator, CFStringRef theParentID, CFStringRef thePermanentID, CFStringRef theFolderName, CFStringRef theLibraryTitle, CFStringRef theNotes ) {
	
	MMErrorRef eref = NULL;
	CFURLRef url = NULL;
	
	if( theParentID != NULL && thePermanentID != NULL && theFolderName != NULL ) {
		url = MMU_URLCreateFolderInParent( theAllocator, theParentID, theFolderName, &eref );
	}
	else {
		eref = MMErrorCreate( theAllocator, kMMErrorDomainOSStatus, MMErrorGetParamErr(), CFSTR( "%s" ), __func__ );
	}
	if( url != NULL ) {
		
		MMUUIDRef ldid = MMLibraryDescriptionCreateMinimal( theAllocator, theParentID, thePermanentID, theLibraryTitle, &eref );
		
		if( ldid != NULL ) {
			
			/*! kMMFileInfoStoreAll will create POSIX path, HFS path, and CFURLBookmark data.
			This is a lot of bytes that aren't needed. The paths can be generated dynamically from the URL string
			so the only one you might need to use is bookmark data -- and only in the case where alias-like tracking is needed.
			*/
			
			//CFNumberRef storage_types = MMU_NumberCreateWithCFIndex( theAllocator, kMMFileInfoStoreAll );
			CFNumberRef storage_types = MMU_NumberCreateWithCFIndex( theAllocator, kMMFileInfoStoreMinimal );
			
			/* MMLibraryDescriptionSetFolderInfoWithURL can take a NULL 'theFlags' parameter so we don't care if CFNumberCreate() fails. */
			eref = MMLibraryDescriptionSetFolderInfoWithURL( theAllocator, ldid, url, storage_types );
			MMU_SafeRelease( storage_types, NULL );
			
			if( eref == NULL ) {
				if( theNotes != NULL ) {
					eref = MMLibraryDescriptionPropertySetValue( theAllocator, ldid, kMMKeyNotes, theNotes );
				}
			}
#if 0
			if( eref == NULL ) {
				
				/* Add alias data for a file URL -- Generates many bytes so only use when absolutely necessary. */
				CFDataRef data = MMAliasDataCreateWithFileURL( theAllocator, url, &eref );
				
				if( data != NULL ) {
					eref = MMLibraryDescriptionFolderInfoSetValue( theAllocator, ldid, kMMFileInfoKeyAliasData, data );
					CFRelease( data );
				}
			}
#endif
			if( eref == NULL ) {
				eref = MMDocumentRegisterLibrary( theAllocator, ldid, NULL );
			}
			CFRelease( ldid );
		}
		CFRelease( url );
	}
	
	return eref;
}


CFMutableStringRef MMTestFuncs( CFAllocatorRef theAllocator ) {
	
	int err = 0;
	CFIndex count = 0;
	CFRange range;
	MMErrorRef eref = NULL;
	CFStringRef cfst = NULL;
	CFURLRef url = NULL;
	CFMutableArrayRef cfma = NULL;
	AEDesc desc = {typeNull, NULL};
	CFMutableStringRef output = CFStringCreateMutable( theAllocator, 0 );
	CFBooleanRef kInstallExamples = kCFBooleanTrue;
	CFMutableDictionaryRef document_options = NULL;
	
	if( output == NULL ) {return output;}
	CFStringAppendFormat( output, NULL, CFSTR( "ModuleManager.framework BETA version: %@\n" ), kMMDocumentVersion );
	
	document_options = CFDictionaryCreateMutable( theAllocator, 0, &kCFCopyStringDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks );
	if( document_options == NULL ) {return output;};
	
	CFDictionarySetValue( document_options, kMMDocumentOptionInstallExamples, kInstallExamples );
	eref = ModuleManagerLoadDocument( theAllocator, document_options );
	if( eref != NULL ) {
		MMErrorShow( eref, CFSTR( "ModuleManagerLoadDocument:" ) );
		CFRelease( eref );
		CFRelease( document_options );
		return output;
	}
	
	CFStringAppend( output, CFSTR( "\nMMDocumentCopyLibraryIDs:\n" ) );
	cfma = MMDocumentCopyLibraryIDs( theAllocator, &eref );
	if( cfma != NULL ) {
		count = CFArrayGetCount( cfma );
		range = CFRangeMake( 0, count );
		CFArrayApplyFunction( cfma, range, ShowSomeValues, (void *)output );
		CFRelease( cfma );
	}
	
	CFStringAppend( output, CFSTR( "\nWhen addressing a library by title, always check for duplicates.\nMMDocumentCopyLibraryIDsWithTitle:\n" ) );
	cfst = MMLibraryPropertyGetValue( theAllocator, kMMUserLibraryID, kMMKeyTitle, &eref );
	if( cfst != NULL ) {
		cfma = MMDocumentCopyLibraryIDsWithTitle( theAllocator, cfst, &eref );
		if( cfma != NULL ) {
			count = CFArrayGetCount( cfma );
			range = CFRangeMake( 0, count );
			CFArrayApplyFunction( cfma, range, GetTitles, (void *)output );
			CFRelease( cfma );
		}
	}
	
	CFStringAppend( output, CFSTR( "\nURL alias desc:\n" ) );
	url = ModuleManagerGetModulesFolderURLForUserDomain();
	err = MMCreateAliasDescWithFileURL( url, &desc );
	if( err == noErr ) {
		cfst = AEPrint( &desc );
		if( cfst != NULL ) {
			CFStringAppend( output, cfst );
			CFRelease( cfst );
		}
	}
	(void)AEDisposeDesc( &desc );
	
	
	CFStringAppend( output, CFSTR( "\nURLString text desc:\n" ) );
	err = MMCreateUTF8TextDescWithURL( url, &desc );
	if( err == noErr ) {
		cfst = AEPrint( &desc );
		if( cfst != NULL ) {
			CFStringAppend( output, cfst );
			CFRelease( cfst );
		}
	}
	(void)AEDisposeDesc( &desc );
	
	CFStringAppend( output, CFSTR( "\nURL POSIXPath desc:\n" ) );
	err = MMCreatePOSIXPathDescWithURL( url, &desc );
	if( err == noErr ) {
		cfst = AEPrint( &desc );
		if( cfst != NULL ) {
			CFStringAppend( output, cfst );
			CFRelease( cfst );
		}
	}
	(void)AEDisposeDesc( &desc );
	
	
	CFStringAppend( output, CFSTR( "\nTry adding new libraries:\n" ) );
	
	eref = CreateEmptyLibrary( theAllocator );
	MMErrorShow( eref, CFSTR( "EmptyLibrary" ) );MMU_SafeRelease( eref, NULL );
	
	eref = CreateNewLibraryAndFolder( theAllocator, kMMUserLibraryID, kPermanentLibraryIDFileSystem, CFSTR( "FileSystem" ), CFSTR( "File System Utilities" ), CFSTR( "File system utilities using System Events" ) );
	MMErrorShow( eref, CFSTR( "FileSystem" ) );MMU_SafeRelease( eref, NULL );
	
	eref = CreateNewLibraryAndFolder( theAllocator, kPermanentLibraryIDFileSystem, kPermanentLibraryIDFiles, CFSTR( "Files" ), CFSTR( "File Utilities" ), CFSTR( "File utilities using System Events" ) );
	MMErrorShow( eref, CFSTR( "Files" ) );MMU_SafeRelease( eref, NULL );
	
	eref = CreateNewLibraryAndFolder( theAllocator, kPermanentLibraryIDFileSystem, kPermanentLibraryIDFolders, CFSTR( "Folders" ), CFSTR( "Folder Utilities" ), CFSTR( "Folder utilities using System Events" ) );
	MMErrorShow( eref, CFSTR( "Folders" ) );MMU_SafeRelease( eref, NULL );
	
	eref = CreateNewLibraryAndFolder( theAllocator, kPermanentLibraryIDFileSystem, kPermanentLibraryIDDisks, CFSTR( "Disks" ), CFSTR( "Disk Utilities" ), CFSTR( "Disks utilities without System Events" ) );
	MMErrorShow( eref, CFSTR( "Disks" ) );MMU_SafeRelease( eref, NULL );
	
	eref = CreateNewLibraryAndFolder( theAllocator, kPermanentLibraryIDDisks, kPermanentLibraryIDDisksSub10, CFSTR( "DisksSub10" ), CFSTR( "Disk Utilities 10" ), NULL );
	MMErrorShow( eref, CFSTR( "Disks" ) );MMU_SafeRelease( eref, NULL );
	
	eref = CreateNewLibraryAndFolder( theAllocator, kPermanentLibraryIDDisksSub10, kPermanentLibraryIDDisksSub20, CFSTR( "DisksSub20" ), CFSTR( "Disk Utilities 20" ), NULL );
	MMErrorShow( eref, CFSTR( "Disks" ) );MMU_SafeRelease( eref, NULL );
	
	eref = CreateNewLibraryAndFolder( theAllocator, kPermanentLibraryIDDisksSub10, kPermanentLibraryIDDisksSub21, CFSTR( "DisksSub21" ), CFSTR( "Disk Utilities 21" ), NULL );
	MMErrorShow( eref, CFSTR( "Disks" ) );MMU_SafeRelease( eref, NULL );
	
	
	CFStringAppend( output, CFSTR( "\nModuleManagerCopyLibraryIDs after adding new libraries:\n" ) );
	cfma = MMDocumentCopyLibraryIDs( theAllocator, &eref );
	if( cfma != NULL ) {
		count = CFArrayGetCount( cfma );
		range = CFRangeMake( 0, count );
		CFArrayApplyFunction( cfma, range, ShowSomeValues, (void *)output );
		CFRelease( cfma );
	}
	
	CFStringAppend( output, CFSTR( "\nCheck which library titles contain substring: \"utilities\":\n" ) );
	cfma = MMDocumentCopyLibraryIDsWithTitleContainingSubstring( theAllocator, CFSTR( "utilities" ), &eref );
	if( cfma != NULL ) {
		count = CFArrayGetCount( cfma );
		range = CFRangeMake( 0, count );
		CFArrayApplyFunction( cfma, range, ShowSomeValues, (void *)output );
		CFRelease( cfma );
	}
	
	CFStringAppend( output, CFSTR( "\nTry deleting some libraries: ...\n" ) );
	eref = MMDocumentRemoveLibrary( theAllocator, kPermanentLibraryIDFolders );
	CFStringAppend( output, CFSTR( "\t--'Folders' library deleted\n" ) );
	eref = MMDocumentRemoveLibrary( theAllocator, kPermanentLibraryIDDisks );
	CFStringAppend( output, CFSTR( "\t--'Disks' library deleted\n" ) );
	
	CFStringAppend( output, CFSTR( "... MMDocumentCopyLibraryIDs after deleting some libraries:\n" ) );
	cfma = MMDocumentCopyLibraryIDs( theAllocator, &eref );
	if( cfma != NULL ) {
		count = CFArrayGetCount( cfma );
		range = CFRangeMake( 0, count );
		CFArrayApplyFunction( cfma, range, ShowSomeValues, (void *)output );
		CFRelease( cfma );
	}
	
	
	CFStringAppend( output, CFSTR( "\nSet User SearchConfiguration title:\n" ) );
	CFStringRef unam = CSCopyUserName( false );
	CFStringRef isot = CurrentImproperlyFormattedISODate( theAllocator );
	if( unam != NULL && isot != NULL ) {
		cfst = CFStringCreateWithFormat( theAllocator, NULL, CFSTR( "%@'s custom search config: %@" ), unam, isot );
		if( cfst != NULL ) {
			eref = MMSearchConfigurationSetValue( theAllocator, kMMSearchConfigurationIDUser, kMMKeyTitle, cfst );
			CFRelease( cfst );
		}
		cfst = MMSearchConfigurationGetValue( theAllocator, kMMSearchConfigurationIDUser, kMMKeyTitle, NULL );
		if( cfst != NULL ) {
			CFStringAppendFormat( output, NULL, CFSTR( "User SearchConfiguration title is now: \"%@\"\n\n" ), cfst );
		}
	}
	MMU_SafeRelease( unam, NULL );
	MMU_SafeRelease( isot, NULL );	
	
	cfma = CFArrayCreateMutable( theAllocator, 0, &kCFTypeArrayCallBacks );
	if( cfma != NULL ) {
		CFArrayAppendValue( cfma, kMMUserLibraryID );
		CFArrayAppendValue( cfma, kPermanentLibraryIDFiles );
		eref = MMSearchConfigurationDescriptionCreate( theAllocator, CFSTR( "User + Files" ), kPermanentSearchConfigUserPlusFiles, cfma );
		if( eref != NULL ) {
			MMErrorShow( eref, CFSTR( "MMSearchConfigurationDescriptionCreate" ) );
			MMU_SafeRelease( eref, NULL );
		}
		if( eref == NULL ) {
			CFStringAppendFormat( output, NULL, CFSTR( "Created new search configuration description with ID: %@.\n" ), kPermanentSearchConfigUserPlusFiles );
			eref = MMDocumentRegisterSearchConfiguration( theAllocator, kPermanentSearchConfigUserPlusFiles );
			if( eref == NULL ) {
				CFStringAppend( output, CFSTR( "Added new search configuration description successfully.\n" ) );
				CFStringAppend( output, CFSTR( "Current search configuration library ids are:\n" ) );
				count = CFArrayGetCount( cfma );
				range = CFRangeMake( 0, count );
				CFArrayApplyFunction( cfma, range, ShowUUIDArray, (void *)output );
				eref = MMDocumentSetSearchConfigurationLastSelected( theAllocator, kPermanentSearchConfigUserPlusFiles );
			}
			else {
				MMErrorShow( eref, CFSTR( "MMDocumentSetSearchConfigurationLastSelected" ) );
				CFRelease( eref );
			}
		}
		MMU_SafeRelease( eref, NULL );
		CFRelease( cfma );
	}
	
#if 0
	/* Check sc delete */
	CFStringAppendFormat( output, NULL, CFSTR( "Deleting search configuration ID: %@.\n" ), kPermanentSearchConfigUserPlusFiles );
	eref = MMDocumentRemoveSearchConfiguration( theAllocator, kPermanentSearchConfigUserPlusFiles );
	if( eref == NULL ) {
		
		MMUUIDRef last_selected = NULL;
		
		CFStringAppend( output, CFSTR( "Deletion successful.\n" ) );
		last_selected = MMDocumentGetSearchConfigurationLastSelected( theAllocator );
		if( last_selected != NULL ) {
			cfst = MMSearchConfigurationGetValue( theAllocator, kMMSearchConfigurationIDUser, kMMKeyTitle, NULL );
			if( cfst != NULL ) {
				CFStringAppendFormat( output, NULL, CFSTR( "Current active search configuration is: %@.\n" ), cfst );
			}
		}
	}
	else {
		MMErrorShow( eref, CFSTR( "MMDocumentRemoveSearchConfiguration failed." ) );
		CFRelease( eref );
	}
#endif
	
	/* Example files */
	if( 1 /*kInstallExamples == kCFBooleanTrue*/ ) {
		
		/* This script is at a known location when the examples are installed. */
		CFStringRef snam = CFSTR( "SE_ListDesktopFolder.scpt" );
		CFURLRef surl = CFURLCreateCopyAppendingPathComponent( theAllocator, ModuleManagerGetModulesFolderURLForUserDomain(), snam, false );
		MMUUIDRef uuid = NULL; /* This id will be associated with the URL. Keep it around for faster access if needed. */
#if 0
		/* If you only want to register reachable URLs use this. */
		if( ! MMU_URLTargetExists( theAllocator, surl ) ) {
			MMU_SafeRelease( surl, NULL );
			/* jump out of this 'if' */
		}
#endif
		CFStringAppendFormat( output, NULL, CFSTR( "\nAttempting to register '%@' into the user library\n" ), snam );
		uuid = MMLibraryRegisterURL( theAllocator, kMMUserLibraryID, surl, &eref );
		if( uuid != NULL ) {
			CFStringAppendFormat( output, NULL, CFSTR( "Registering '%@' successful -> %@\n" ), snam, uuid );
			/* We're keeping this id because we want to register it into another library. Otherwise release it. */
		}
		else {
			MMErrorShow( eref, CFSTR( "Registration into user library failed" ) );
			CFRelease( eref );
		}
		
		CFStringAppendFormat( output, NULL, CFSTR( "\nAttempting to register '%@' into the Files library using the same ID.\n" ), snam );
		eref = MMLibraryRegisterFileID( theAllocator, kPermanentLibraryIDFiles, uuid );
		if( eref == NULL ) {
			CFStringAppendFormat( output, NULL, CFSTR( "Registering '%@' successful -> %@\n" ), snam, uuid );
		}
		else {
			MMErrorShow( eref, CFSTR( "Registration into Files library failed" ) );
			CFRelease( eref );
		}
		
		CFStringAppend( output, CFSTR( "\nShow the files in the user library:\n" ) );
		cfma = MMLibraryGetFiles( theAllocator, kMMUserLibraryID, &eref );
		if( cfma != NULL ) {
			count = CFArrayGetCount( cfma );
			if( count > 0 ) {
				range = CFRangeMake( 0, count );
				CFArrayApplyFunction( cfma, range, ShowStringArray, (void *)output );
			}
		}
		
		CFStringAppend( output, CFSTR( "\nShow the files in the Files library:\n" ) );
		cfma = MMLibraryGetFiles( theAllocator, kPermanentLibraryIDFiles, &eref );
		if( cfma != NULL ) {
			count = CFArrayGetCount( cfma );
			if( count > 0 ) {
				range = CFRangeMake( 0, count );
				CFArrayApplyFunction( cfma, range, ShowStringArray, (void *)output );
			}
		}
#if 0
		/* Now remove a file */
		CFStringAppend( output, CFSTR( "\nRemove a file from the user library:\n" ) );
		
		/* This is slower: eref = MMLibraryUnregisterURL( theAllocator, kMMUserLibraryID, surl ); */
		eref = MMLibraryUnregisterFileID( theAllocator, kMMUserLibraryID, uuid );
		if( eref != NULL ) {
			MMErrorShow( eref, CFSTR( "MMLibraryUnregisterFileID failed." ) );
			CFRelease( eref );
		}
		
		CFStringAppend( output, CFSTR( "\nShow the files in the user library:\n" ) );
		cfma = MMLibraryGetFiles( theAllocator, kMMUserLibraryID, NULL );
		if( cfma != NULL ) {
			count = CFArrayGetCount( cfma );
			if( count > 0 ) {
				range = CFRangeMake( 0, count );
				CFArrayApplyFunction( cfma, range, ShowStringArray, (void *)output );
			}
		}
		
		CFStringAppend( output, CFSTR( "\nShow the files in the Files library:\n" ) );
		cfma = MMLibraryGetFiles( theAllocator, kPermanentLibraryIDFiles, NULL );
		if( cfma != NULL ) {
			count = CFArrayGetCount( cfma );
			if( count > 0 ) {
				range = CFRangeMake( 0, count );
				CFArrayApplyFunction( cfma, range, ShowStringArray, (void *)output );
			}
		}
#endif
		if( surl != NULL ) {
			CFRelease( surl );
		}
		if( uuid != NULL ) {
			CFRelease( uuid );
		}
	}
	if( eref == NULL ) {
		/* Put some folders with files in ~/Library/Scripts/Modules.x.x.x/ and observe the document entries. */
		eref = MMLibraryRegisterSubfolders( theAllocator, kMMUserLibraryID );
	}
	
	if( document_options != NULL ) {
		CFRelease( document_options );
	}
	if( eref == NULL ) {
		eref = ModuleManagerSaveDocument( theAllocator );
	}
	if( eref != NULL ) {
		MMErrorShow( eref, NULL );
		CFRelease( eref );
	}
	
	return output;
}

