/* MMTestingMain.c */

#include <Carbon/Carbon.h>
#include "MMTestFuncs.h"

static IBNibRef gNibRef;
static OSStatus AppEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void *inRefcon );
static OSStatus HandleNew( void );
static OSStatus WindowEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void *inRefcon );
static OSStatus CreateNewLibraryFolder( HIViewRef theView );


OSStatus CreateNewLibraryFolder( HIViewRef theView ) {
	
	int err = 0;
	CFMutableStringRef text = MMTestFuncs( kCFAllocatorDefault );
	
	if( text != NULL ) {
		err = HIViewSetText( theView, text );
		CFRelease( text );
	}
	
	return err;
}


OSStatus AppEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void *inRefCon ) {
	
	#pragma unused( inCaller, inRefCon )
	
	OSStatus result = eventNotHandledErr;
	
	switch( GetEventClass( inEvent ) ) {
		case kEventClassCommand: {
			
			HICommandExtended cmd;
			
			GetEventParameter( inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof( cmd ), NULL, &cmd );
			switch( GetEventKind( inEvent ) ) {
				case kEventCommandProcess:
					switch( cmd.commandID ) {
						case kHICommandNew: {
							result = HandleNew();
							break;
						}
						default: {
							break;
						}
					}
					break;
			}
			break;
		}
		default: {
			break;
		}
	}

	return result;
}


OSStatus HandleNew( void ) {
	
	OSStatus err = noErr;
	WindowRef window = NULL;
	HIViewRef view = NULL;
	HIViewID hvid = {'TViw', 0};
	static const EventTypeSpec kWindowEvents[] = {
		{kEventClassCommand, kEventCommandProcess},
	};
	
	err = CreateWindowFromNib( gNibRef, CFSTR( "MainWindow" ), &window );
	err = HIViewFindByID( HIViewGetRoot( window ), hvid, &view );
	if( view != NULL ) {
		err = CreateNewLibraryFolder( view );
	}
#if 0
	HIRect wrct;
	err = HIWindowGetAvailablePositioningBounds( kCGNullDirectDisplay, kHICoordSpace72DPIGlobal, &wrct );
	if( err == noErr ) {
		Rect rect = {(short)wrct.origin.y, (short)wrct.origin.x, (short)(wrct.size.height + wrct.origin.y), (short)(wrct.size.width - wrct.origin.x)};
		err = SetWindowBounds( window, kWindowStructureRgn, &rect );
	}
#endif
	err = InstallWindowEventHandler( window, NewEventHandlerUPP( WindowEventHandler ), GetEventTypeCount( kWindowEvents ), kWindowEvents, window, NULL );
	ShowWindow( window );
	
	return err;
}


OSStatus WindowEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void *inRefCon ) {
	
	#pragma unused( inCaller, inRefCon )
	
	OSStatus err = eventNotHandledErr;
	
	switch( GetEventClass( inEvent ) ) {
		case kEventClassCommand: {
			
			HICommandExtended xcmd;
			
			GetEventParameter( inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof( xcmd ), NULL, &xcmd );
			switch( GetEventKind( inEvent ) ) {
				case kEventCommandProcess: {
					switch( xcmd.commandID ) {
						/* Add your own command-handling cases here. */
						default: {
							break;
						}
					}
					break;
				}
			
			}
			break;
		}
		default: {
			break;
		}
	}
	
	return err;
}


int main( int argc, char *argv[] ) {
	
	#pragma unused( argc, argv )
	
	OSStatus err = noErr;
	static const EventTypeSpec kAppEvents[] = {{kEventClassCommand, kEventCommandProcess}};
	
	err = CreateNibReference( CFSTR( "main" ), &gNibRef );
	err = SetMenuBarFromNib( gNibRef, CFSTR( "MenuBar" ) );
	err = InstallApplicationEventHandler( NewEventHandlerUPP( AppEventHandler ), GetEventTypeCount( kAppEvents ), kAppEvents, 0, NULL );
	err = HandleNew();
	RunApplicationEventLoop();
	
	return err;
}

