/* ModMan.c */

#include <Carbon/Carbon.h>
#include "MMTestFuncs.h"

enum {
	kModMan_EventClass_ModuleManager = 'Mdul',
	kModMan_EventID_Test = 'Test',
};

struct ModMan_AEEventHandlerInfo {
	FourCharCode evcl;
	FourCharCode evid;
	AEEventHandlerProcPtr proc;
};
typedef struct ModMan_AEEventHandlerInfo ModMan_AEEventHandlerInfo;

static AEEventHandlerUPP *lgModMan_EventUPPs;
CFBundleRef gModMan_OSAXBundleRef = NULL;

extern OSErr SAInitialize( CFBundleRef theBundle );
extern void SATerminate( void );
extern Boolean SAIsBusy( void );


static OSErr ModMan_EventHandlersInstall( void );
static void ModMan_EventHandlersRemove( void );
extern OSErr ModMan_EventHandler_ModuleManagerTest( const AppleEvent *theAppleEvent, AppleEvent *outReply, SRefCon ioRefCon );

const ModMan_AEEventHandlerInfo lgModMan_EventInfo[] = {
	{kModMan_EventClass_ModuleManager, kModMan_EventID_Test, ModMan_EventHandler_ModuleManagerTest},
};
const int kModMan_EventHandlerCount = (sizeof( lgModMan_EventInfo ) / sizeof( ModMan_AEEventHandlerInfo ));


OSErr SAInitialize( CFBundleRef theBundle ) {
	
	OSErr err = noErr;
CFShow( theBundle );
	gModMan_OSAXBundleRef = theBundle;
	lgModMan_EventUPPs = (AEEventHandlerUPP *)malloc( sizeof( AEEventHandlerUPP ) * kModMan_EventHandlerCount );
	err = ModMan_EventHandlersInstall();
	
	return err;
}


void SATerminate( void ) {
	free( lgModMan_EventUPPs );
	ModMan_EventHandlersRemove();
}


Boolean SAIsBusy( void ) {
	return false;
}


OSErr ModMan_EventHandlersInstall( void ) {
	
	OSErr err = noErr;
	
	for( size_t i = 0; i < kModMan_EventHandlerCount; i++ ) {
		if( (lgModMan_EventUPPs[i] = NewAEEventHandlerUPP( lgModMan_EventInfo[i].proc)) != NULL ) {
			err = AEInstallEventHandler( lgModMan_EventInfo[i].evcl, lgModMan_EventInfo[i].evid, lgModMan_EventUPPs[i], 0, true );
		}
		else {
			err = memFullErr;
		}
		if( err != noErr ) {
			/* Call the termination function ourselves, because the loader won't once we fail. */
			SATerminate();
			return err;
		}
	}
	
	return err; 
}


void ModMan_EventHandlersRemove( void ) {
	
	OSErr err = noErr;
	size_t i = 0;
	
	for( i = 0; i < kModMan_EventHandlerCount; i++ ) {
		if( lgModMan_EventUPPs[i] && (err = AERemoveEventHandler( lgModMan_EventInfo[i].evcl, lgModMan_EventInfo[i].evid, lgModMan_EventUPPs[i], true)) == noErr ) {
			DisposeAEEventHandlerUPP( lgModMan_EventUPPs[i] );
			lgModMan_EventUPPs[i] = NULL;
		}
	}
}


OSErr ModMan_EventHandler_ModuleManagerTest( const AppleEvent *theAppleEvent, AppleEvent *outReply, SRefCon ioRefCon ) {
	
	#pragma unused( ioRefCon )
	
	OSErr err = noErr;
	AEEventClass evcl = 0;
	AEEventID evid = 0;
	CFMutableStringRef mute = NULL;
	AEDesc desc = {typeNull, NULL};
	
	err = AEGetAttributePtr( theAppleEvent, keyEventClassAttr, typeType, NULL, &evcl, sizeof( evcl ), NULL );
	err = AEGetAttributePtr( theAppleEvent, keyEventIDAttr, typeType, NULL, &evid, sizeof( evid ), NULL );
	
	switch( evcl ) {
		case kModMan_EventClass_ModuleManager: {
			switch( evid ) {
				case kModMan_EventID_Test: {
					mute = MMTestFuncs( kCFAllocatorDefault );
					if( mute != NULL ) {
						err = MMCreateUTF8DescWithCFString( mute, &desc );
						if( err == noErr ) {
							err = AEPutParamDesc( outReply, keyAEResult, &desc );
						}
						CFRelease( mute );
					}
					break;
				}
			}
			break;
		}
	}
	(void)AEDisposeDesc( &desc );
	
	return err;
}
